"use strict";
var _a;
Object.defineProperty(exports, "__esModule", { value: true });
var ts = require("typescript");
var path = require("path");
var ts_morph_1 = require("ts-morph");
var CUSTOM_ELEMENT_NAMES = {
    API_SUCCESS_INTERFACE: 'apisuccessinterface',
    API_PARAM_INTERFACE: 'apiparaminterface',
    API_QUERY_INTERFACE: 'apiqueryinterface',
    API_BODY_INTERFACE: 'apibodyinterface'
};
var APIDOC_ELEMENT_BY_CUSTOM_ELEMENT_NAME = (_a = {},
    _a[CUSTOM_ELEMENT_NAMES.API_SUCCESS_INTERFACE] = 'apiSuccess',
    _a[CUSTOM_ELEMENT_NAMES.API_PARAM_INTERFACE] = 'apiParam',
    _a[CUSTOM_ELEMENT_NAMES.API_QUERY_INTERFACE] = 'apiQuery',
    _a[CUSTOM_ELEMENT_NAMES.API_BODY_INTERFACE] = 'apiBody',
    _a);
var definitionFilesAddedByUser = {};
var Apidoc;
(function (Apidoc) {
    var AvailableHook;
    (function (AvailableHook) {
        AvailableHook["parser-find-elements"] = "parser-find-elements";
    })(AvailableHook = Apidoc.AvailableHook || (Apidoc.AvailableHook = {}));
})(Apidoc || (Apidoc = {}));
var ast = new ts_morph_1.Project();
function init(app) {
    app.addHook(Apidoc.AvailableHook['parser-find-elements'], parseElements.bind(app), 200);
}
exports.init = init;
function parseElements(elements, element, block, filename) {
    if (!Object.values(CUSTOM_ELEMENT_NAMES).includes(element.name))
        return;
    elements.pop();
    var newElements = [];
    var values = parse(element.content, element.name);
    if (!values) {
        this.log.warn("Could not find parse values of element: " + element.content);
        return;
    }
    var namedInterface = values.interface.trim();
    var interfacePath = values.path ? path.resolve(path.dirname(filename), values.path.trim()) : filename;
    var parentNamespace = parseDefinitionFiles.call(this, interfacePath);
    var _a = extractNamespace.call(this, parentNamespace, namedInterface), namespace = _a.namespace, leafName = _a.leafName;
    if (isNativeType(leafName)) {
        parseNative(elements, newElements, interfacePath, values);
        return;
    }
    var arrayMatch = matchArrayInterface(leafName);
    if (arrayMatch) {
        parseArray.call(this, elements, newElements, values, interfacePath, namespace, arrayMatch);
        return;
    }
    parseInterface.call(this, elements, newElements, values, interfacePath, namespace, leafName);
}
function parseNative(elements, newElements, interfacePath, values) {
    setNativeElements(interfacePath, newElements, values);
    elements.push.apply(elements, newElements);
}
function parseArray(elements, newElements, values, interfacePath, namespace, arrayMatch) {
    var interfaceName = arrayMatch.interface;
    var matchedInterface = getNamespacedInterface(namespace, interfaceName);
    if (!matchedInterface) {
        this.log.warn("Could not find interface \u00AB" + interfaceName + "\u00BB in file \u00AB" + interfacePath + "\u00BB");
        return;
    }
    setArrayElements.call(this, matchedInterface, interfacePath, newElements, values, interfaceName);
    elements.push.apply(elements, newElements);
}
function parseInterface(elements, newElements, values, interfacePath, namespace, leafName) {
    var matchedInterface = getNamespacedInterface(namespace, leafName);
    if (!matchedInterface) {
        this.log.warn("Could not find interface \u00AB" + values.interface + "\u00BB in file \u00AB" + interfacePath + "\u00BB");
        return;
    }
    setInterfaceElements.call(this, matchedInterface, interfacePath, newElements, values);
    elements.push.apply(elements, newElements);
}
var PropType;
(function (PropType) {
    PropType["Enum"] = "Enum";
    PropType["Array"] = "Array";
    PropType["Object"] = "Object";
    PropType["Native"] = "Native";
})(PropType || (PropType = {}));
function parse(content, elementName) {
    if (content.length === 0)
        return null;
    var parseRegExp = /^(?:\((.+?)\)){0,1}\s*\{(.+?)\}\s*(?:([^\s]*))?\s*(?:(.+))?/g;
    var matches = parseRegExp.exec(content);
    if (!matches)
        return null;
    return {
        element: APIDOC_ELEMENT_BY_CUSTOM_ELEMENT_NAME[elementName],
        path: matches[1],
        interface: matches[2],
        field: matches[3] || '',
        description: matches[4] || ''
    };
}
function setArrayElements(matchedInterface, filename, newElements, values, interfaceName) {
    var field = values.field || getDecapitalized(interfaceName);
    var description = values.description || field;
    newElements.push(getApiElement(values.element, "{Object[]} " + field + " " + description));
    setInterfaceElements.call(this, matchedInterface, filename, newElements, values, field);
}
function setInterfaceElements(matchedInterface, filename, newElements, values, inttype) {
    var _this = this;
    extendInterface.call(this, matchedInterface, filename, newElements, values, inttype);
    matchedInterface.getProperties().forEach(function (prop) {
        var typeDef = inttype ? inttype + "." + prop.getName() : prop.getName();
        var documentationComments = prop.getJsDocs().map(function (node) { return node.getInnerText(); }).join();
        var description = documentationComments
            ? "`" + typeDef + "` - " + documentationComments
            : "`" + typeDef + "`";
        var propTypeName = prop.getType().getText(undefined, ts.TypeFormatFlags.UseAliasDefinedOutsideCurrentScope);
        var typeEnum = getPropTypeEnum(prop);
        var propLabel = getPropLabel(typeEnum, propTypeName);
        newElements.push(getApiElement(values.element, "{" + propLabel + "} " + typeDef + " " + description));
        if ([PropType.Object, PropType.Array].includes(typeEnum)) {
            var typeInterface = getInterface.call(_this, filename, propTypeName);
            var arrayType = typeEnum === PropType.Array && prop.getType().getArrayElementType();
            var objectProperties = arrayType
                ? arrayType.getProperties()
                : prop.getType().getProperties();
            if (typeInterface) {
                setInterfaceElements.call(_this, typeInterface, filename, newElements, values, typeDef);
            }
            else {
                setObjectElements.call(_this, objectProperties, filename, newElements, values, typeDef);
            }
        }
    });
}
function setNativeElements(filename, newElements, values) {
    var propLabel = getCapitalized(values.interface);
    newElements.push(getApiElement(values.element, "{" + propLabel + "} " + values.field + " " + values.description));
    return;
}
function setObjectElements(properties, filename, newElements, values, typeDef) {
    var _this = this;
    properties.forEach(function (property) {
        var valueDeclaration = property.getValueDeclaration();
        if (!valueDeclaration)
            return;
        var propName = property.getName();
        var typeDefLabel = typeDef + "." + propName;
        var propType = valueDeclaration.getType().getText(valueDeclaration);
        var isUserDefinedProperty = isUserDefinedSymbol(property.compilerSymbol);
        if (!isUserDefinedProperty)
            return;
        var documentationComments = '';
        try {
            documentationComments = property.compilerSymbol.getDocumentationComment(undefined).map(function (node) { return node.text; }).join();
        }
        catch (err) {
            _this.log.warn(err);
        }
        var desc = documentationComments
            ? "`" + typeDef + "." + propName + "` - " + documentationComments
            : "`" + typeDef + "." + propName + "`";
        if (isNativeType(propType)) {
            newElements.push(getApiElement(values.element, "{" + getCapitalized(propType) + "} " + typeDefLabel + " " + desc));
            return;
        }
        var isEnum = valueDeclaration.getType().isEnum();
        if (isEnum) {
            newElements.push(getApiElement(values.element, "{Enum} " + typeDefLabel + " " + desc));
            return;
        }
        var newElement = getApiElement(values.element, "{Object" + (propType.includes('[]') ? '[]' : '') + "} " + typeDefLabel + " " + desc);
        newElements.push(newElement);
        var typeInterface = getInterface.call(_this, filename, propType);
        if (typeInterface) {
            setInterfaceElements.call(_this, typeInterface, filename, newElements, values, typeDefLabel);
        }
        else {
            var externalFileTypeSymbol = valueDeclaration.getType().getSymbol();
            if (!externalFileTypeSymbol) {
                setObjectElements.call(_this, property.getValueDeclarationOrThrow().getType().getProperties(), filename, newElements, values, typeDef);
                return;
            }
            var externalFileDeclaration = externalFileTypeSymbol.getDeclarations()[0];
            var externalFileInterface = externalFileDeclaration.getSourceFile().getInterface(propType);
            if (!externalFileInterface) {
                setObjectElements.call(_this, property.getValueDeclarationOrThrow().getType().getProperties(), filename, newElements, values, typeDefLabel);
                return;
            }
            setObjectElements.call(_this, externalFileInterface.getType().getProperties(), filename, newElements, values, typeDefLabel);
        }
    });
}
function extendInterface(matchedInterface, interfacePath, newElements, values, inttype) {
    for (var _i = 0, _a = matchedInterface.getExtends(); _i < _a.length; _i++) {
        var extendedInterface = _a[_i];
        var extendedInterfaceName = extendedInterface.compilerNode.expression.getText();
        var parentNamespace = matchedInterface.getParentNamespace() || parseDefinitionFiles.call(this, interfacePath);
        var _b = extractNamespace.call(this, parentNamespace, extendedInterfaceName), namespace = _b.namespace, leafName = _b.leafName;
        var matchedExtendedInterface = getNamespacedInterface.call(this, namespace, leafName);
        if (!matchedExtendedInterface) {
            this.log.warn("Could not find interface to be extended " + extendedInterfaceName);
            return;
        }
        extendInterface.call(this, matchedExtendedInterface, interfacePath, newElements, values);
        setInterfaceElements.call(this, matchedExtendedInterface, interfacePath, newElements, values, inttype);
    }
}
function getApiElement(element, param) {
    return {
        content: param + "\n",
        name: element.toLowerCase(),
        source: "@" + element + " " + param + "\n",
        sourceName: element
    };
}
function parseDefinitionFiles(interfacePath) {
    var interfaceFile = ast.addExistingSourceFile(interfacePath);
    if (!interfaceFile)
        return;
    trackUserAddedDefinitionFile(interfaceFile);
    for (var _i = 0, _a = ast.resolveSourceFileDependencies(); _i < _a.length; _i++) {
        var file = _a[_i];
        trackUserAddedDefinitionFile(file);
    }
    return interfaceFile;
}
function extractNamespace(rootNamespace, interfaceName) {
    var _this = this;
    var arrayMatch = matchArrayInterface(interfaceName);
    interfaceName = arrayMatch
        ? arrayMatch.interface + '[]'
        : interfaceName;
    var isNamespaced = interfaceName.match(/(?:[a-zA-Z0-9_]\.)*[a-zA-Z0-9_]\./i);
    var nameSegments = isNamespaced
        ? interfaceName.split('.')
        : [interfaceName];
    var namespaces = nameSegments.slice(0, -1);
    var leafName = nameSegments[nameSegments.length - 1];
    var namespace = namespaces.reduce(function (parent, name) {
        if (!parent)
            return;
        var namespace = parent.getNamespace(name);
        if (!namespace)
            _this.log.warn("Could not find namespace " + name + " in root namespace in file at " + rootNamespace.getSourceFile().getFilePath());
        return namespace;
    }, rootNamespace);
    return {
        namespace: namespace,
        leafName: leafName
    };
}
function getNamespacedInterface(namespace, interfaceName) {
    return namespace.getInterface(interfaceName);
}
function getInterface(interfacePath, interfaceName) {
    var interfaceFile = parseDefinitionFiles(interfacePath);
    var _a = extractNamespace.call(this, interfaceFile, interfaceName), namespace = _a.namespace, leafName = _a.leafName;
    return getNamespacedInterface.call(this, namespace, leafName);
}
function trackUserAddedDefinitionFile(file) {
    definitionFilesAddedByUser[file.getFilePath()] = true;
}
function getCapitalized(text) {
    return text.charAt(0).toUpperCase() + text.slice(1).toLowerCase();
}
function getDecapitalized(text) {
    return text.charAt(0).toLowerCase() + text.slice(1);
}
function isNativeType(propType) {
    var nativeTypes = ['boolean', 'Boolean', 'string', 'String', 'number', 'Number', 'Date', 'any'];
    return nativeTypes.indexOf(propType) >= 0;
}
function getPropTypeEnum(prop) {
    var propType = prop.getType().getText();
    var propTypeIsEnum = prop.getType().isEnum();
    var propTypeIsObject = !propTypeIsEnum && !isNativeType(propType);
    var propTypeIsArray = propTypeIsObject && propType.includes('[]');
    if (propTypeIsArray)
        return PropType.Array;
    if (propTypeIsObject)
        return PropType.Object;
    if (propTypeIsEnum)
        return PropType.Enum;
    return PropType.Native;
}
function getPropLabel(typeEnum, propTypeName) {
    if (typeEnum === PropType.Array)
        return 'Object[]';
    if (typeEnum === PropType.Object)
        return 'Object';
    if (typeEnum === PropType.Enum)
        return 'Enum';
    return getCapitalized(propTypeName);
}
function matchArrayInterface(interfaceName) {
    var match = interfaceName.match(/^Array<(.*)>$/) || interfaceName.match(/^(.*)\[\]$/);
    if (!match) {
        return null;
    }
    return {
        full: interfaceName,
        interface: match[1]
    };
}
function isUserDefinedSymbol(symbol) {
    var declarationFile = symbol.valueDeclaration.parent.getSourceFile();
    return definitionFilesAddedByUser[declarationFile.fileName];
}
